// BitTestDlg.cpp : implementation file
//

#include "stdafx.h"
#include "BitTest.h"
#include "BitTestDlg.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

const int cx = 128;
const int cy = 128;
const UINT ID_TIMER = 10;

double fps = 0.0;

CBitTestDlg::CBitTestDlg(CWnd* pParent /*=NULL*/)
	: CDialog(CBitTestDlg::IDD, pParent)
{
	//{{AFX_DATA_INIT(CBitTestDlg)
	m_bColor = FALSE;
	//}}AFX_DATA_INIT
	m_hIcon = AfxGetApp()->LoadIcon(IDR_MAINFRAME);

	m_hBitmap = NULL;
	m_pBits = NULL;
	m_BitmapWidth = cx;
	m_BitmapHeight = cy;

	m_hDD = NULL;

	m_hBitmapThread = NULL;
}

void CBitTestDlg::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CBitTestDlg)
	DDX_Control(pDX, IDC_MSG, m_Msg);
	DDX_Check(pDX, IDC_COLOR, m_bColor);
	//}}AFX_DATA_MAP
}

BEGIN_MESSAGE_MAP(CBitTestDlg, CDialog)
	//{{AFX_MSG_MAP(CBitTestDlg)
	ON_WM_PAINT()
	ON_WM_QUERYDRAGICON()
	ON_WM_TIMER()
	ON_BN_CLICKED(IDC_COLOR, OnColor)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

BOOL CBitTestDlg::OnInitDialog()
{
	CDialog::OnInitDialog();

	SetIcon (m_hIcon, TRUE);	// Set big icon
	SetIcon (m_hIcon, FALSE);	// Set small icon

	VERIFY (SetTimer (ID_TIMER, 1000, NULL));
	
	//  
	memset (&m_bmi, 0, sizeof (BITMAPINFO));
	m_bmi.bmiHeader.biSize = sizeof (BITMAPINFOHEADER);
	m_bmi.bmiHeader.biWidth = m_BitmapWidth;
	m_bmi.bmiHeader.biHeight = m_BitmapHeight;
	m_bmi.bmiHeader.biPlanes = 1;
	m_bmi.bmiHeader.biBitCount = 24;		// true-color
	m_bmi.bmiHeader.biCompression = BI_RGB;	// not compressed
	m_bmi.bmiHeader.biXPelsPerMeter = 2880;	// 72 dpi
	m_bmi.bmiHeader.biYPelsPerMeter = 2880;	// 72 dpi
	CDC *pDC = GetDC ();
	ASSERT (pDC != NULL);
	m_hBitmap = CreateDIBSection (pDC->GetSafeHdc (),
		&m_bmi, DIB_RGB_COLORS, (VOID**) &m_pBits, NULL, 0); 
	ASSERT (m_hBitmap != NULL);
	ASSERT (m_pBits != NULL);
	VERIFY (ReleaseDC (pDC));

	m_hDD = DrawDibOpen ();
	ASSERT (m_hDD != NULL);

	unsigned id;
	m_bBitmapThreadStop = FALSE;
	m_hBitmapThread = _beginthreadex (NULL, 0, BitmapThreadFunc, this, 0, &id);
	ASSERT (m_hBitmapThread != 0);
	VERIFY (SetThreadPriority ((void*) m_hBitmapThread, THREAD_PRIORITY_HIGHEST));

	return TRUE;  // return TRUE  unless you set the focus to a control
}

void CBitTestDlg::OnPaint() 
{
	if (IsIconic ()) {
		CPaintDC dc(this); // device context for painting

		SendMessage(WM_ICONERASEBKGND, (WPARAM) dc.GetSafeHdc(), 0);

		// Center icon in client rectangle
		int cxIcon = GetSystemMetrics(SM_CXICON);
		int cyIcon = GetSystemMetrics(SM_CYICON);
		CRect rect;
		GetClientRect(&rect);
		int x = (rect.Width() - cxIcon + 1) / 2;
		int y = (rect.Height() - cyIcon + 1) / 2;

		// Draw the icon
		dc.DrawIcon(x, y, m_hIcon);
	} else {
		CDialog::OnPaint();
	}
}

void CBitTestDlg::DrawBitmap (void)
{
	if (!IsIconic ()) {
		CRect rc;
		GetClientRect (&rc);
		rc.left += ((rc.right - rc.left) - cx) / 2;
		rc.top += ((rc.bottom - rc.top) - cy) / 2;

		CDC *pDC = GetDC ();
		HDC hDC = pDC->GetSafeHdc ();
		
		if (m_bColor) {
			DrawDibBegin (m_hDD, hDC, m_BitmapWidth, m_BitmapHeight,
				&m_bmi.bmiHeader, m_BitmapWidth, m_BitmapHeight,
				DDF_HALFTONE | DDF_ANIMATE);

			DrawDibRealize (m_hDD, hDC, FALSE);

			DrawDibDraw (m_hDD, hDC, rc.left, rc.top, m_BitmapWidth, m_BitmapHeight,
				&m_bmi.bmiHeader, m_pBits, 0, 0, m_BitmapWidth, m_BitmapHeight,
				DDF_SAME_DRAW | DDF_SAME_HDC);
		} else {
			StretchDIBits (hDC, rc.left, rc.top, m_BitmapWidth, m_BitmapHeight,
				0, 0, m_BitmapWidth, m_BitmapHeight, m_pBits, &m_bmi,
				DIB_RGB_COLORS, SRCCOPY);
		}

		ReleaseDC (pDC);
	}
}

HCURSOR CBitTestDlg::OnQueryDragIcon()
{
	return (HCURSOR) m_hIcon;
}

void CBitTestDlg::OnOK() 
{
}

void CBitTestDlg::OnCancel() 
{
	if (m_hBitmapThread != NULL) {
		m_bBitmapThreadStop = TRUE;
		DWORD state = 0;
		do {
			Sleep (100);
			VERIFY (GetExitCodeThread ((void*) m_hBitmapThread, &state));
		} while (state == STILL_ACTIVE);
		VERIFY (CloseHandle ((void*) m_hBitmapThread));
		m_hBitmapThread = NULL;
	}

	VERIFY (KillTimer (ID_TIMER));
	
	if (m_hDD != NULL) {
		DrawDibEnd (m_hDD);
		DrawDibClose (m_hDD);
		m_hDD = NULL;
	}

	VERIFY (DeleteObject (m_hBitmap));

	CDialog::OnCancel();
}

unsigned WINAPI BitmapThreadFunc (void *param)
{
	CBitTestDlg *obj = (CBitTestDlg*) param;
	ASSERT (obj != NULL);

	CDIBFile file1;
	BGR *bits1 = file1.Load ("map1.bmp");
	ASSERT (bits1 != NULL);
	CDIBFile file2;
	BGR *bits2 = file2.Load ("map2.bmp");
	ASSERT (bits2 != NULL);

	LARGE_INTEGER count1, count2, freq;
	QueryPerformanceFrequency (&freq);

	BYTE t = 0;
	while (!obj->m_bBitmapThreadStop) {
		QueryPerformanceCounter (&count1);

		// Draw pixels
		for (int y = 0; y < cy; y++) {
			for (int x = 0; x < cx; x++) {
				DWORD index1 = y * cx + x;
				DWORD index2 = ((y + t) & (cy - 1)) * cx + x;
				obj->m_pBits [index1].R = (BYTE) ((bits1 [index2].R + bits2 [index1].R) >> 1);
				obj->m_pBits [index1].G = (BYTE) ((bits1 [index2].G + bits2 [index1].G) >> 1);
				obj->m_pBits [index1].B = (BYTE) ((bits1 [index2].B + bits2 [index1].B) >> 1);
			}
		}
		t++;
		obj->DrawBitmap ();
		
		QueryPerformanceCounter (&count2);
		fps = 1. / (((double) count2.QuadPart - (double) count1.QuadPart) / (double) freq.QuadPart);
		Sleep (20);
	}

	file1.Clean ();
	file2.Clean ();

	_endthreadex (0);
	return 0;
}

void CBitTestDlg::OnTimer (UINT nIDEvent) 
{
	MSG msg;
	while (::PeekMessage (&msg, m_hWnd, WM_TIMER, WM_TIMER, PM_REMOVE));

	if (nIDEvent == ID_TIMER) {
		CString msg;
		msg.Format ("%7.02f fps", fps);
		m_Msg.SetWindowText (msg);
	} else
		CDialog::OnTimer(nIDEvent);
}

void CBitTestDlg::OnColor() 
{
	UpdateData ();	
}
